# -*- coding: utf-8 -*-
"""ELM Web Scraping - Base class for search engine search"""
import os
import random
import asyncio
import logging
from urllib.parse import quote
from abc import ABC, abstractmethod
from contextlib import asynccontextmanager

from rebrowser_playwright.async_api import async_playwright
from playwright_stealth import StealthConfig

from elm.web.utilities import PWKwargs, clean_search_query, pw_page


logger = logging.getLogger(__name__)


class SearchEngineLinkSearch(ABC):
    """Abstract class to retrieve links for a query using a search engine"""

    _SE_NAME = "<unknown se>"

    async def results(self, *queries, num_results=10):
        """Retrieve links for the first `num_results` of each query

        This function executes a search for each input query and
        returns a list of links corresponding to the top `num_results`.

        Parameters
        ----------
        *queries : str
            One or more queries to search for.
        num_results : int, optional
            Maximum number of top results to retrieve for each query.
            Note that this value can never exceed the number of results
            per page (typically 10). If you pass in a larger value, it
            will be reduced to the number of results per page. There is
            also no guarantee that the search query will return this
            many results - the actual number of results returned is
            determined by the number of results on a page (excluding
            ads). You can, however, use this input to limit the number
            of results returned. By default, ``10``.

        Returns
        -------
        list
            List equal to the length of the input queries, where each
            entry is another list containing no more than `num_results`
            links.
        """
        queries = map(clean_search_query, queries)
        return await self._get_links(queries, num_results)

    async def _get_links(self, queries, num_results):
        """Get links for multiple queries"""
        outer_task_name = asyncio.current_task().get_name()

        searches = [
            asyncio.create_task(
                self._skip_exc_search(query, num_results=num_results),
                name=outer_task_name,
            )
            for query in queries
        ]
        logger.trace("Kicking off %s search for %d queries",
                     self._SE_NAME, len(searches))
        results = await asyncio.gather(*searches)
        logger.trace("Got results for link search:\n%r", results)
        return results

    async def _skip_exc_search(self, query, num_results=10):
        """Perform search while ignoring errors"""
        try:
            return await self._search(query, num_results=num_results)
        except KeyboardInterrupt:
            raise
        except Exception as e:
            msg = "Could not complete search for query=%r\nGot error type: %r"
            logger.exception(msg, query, type(e))
            return []

    async def _move_mouse(self, page):
        """Simulate mouse movement"""
        logger.trace("Moving mouse")
        vp = page.viewport_size or {}
        logger.trace("VP: %r", vp)
        width, height = vp.get('width', 300), vp.get('height', 300)
        min_width, max_width = int(width * 0.1), int(width * 0.9)
        min_height, max_height = int(height * 0.1), int(height * 0.9)
        logger.trace("Moving mouse to random position within: %d-%d, %d-%d",
                     min_width, max_width, min_height, max_height)

        await page.mouse.move(random.randint(min_width, max_width),
                              random.randint(min_height, max_height),
                              steps=10)
        await asyncio.sleep(random.uniform(1.5, 3.5))

    async def _move_and_click(self, page, input_el):
        """Move mouse to an element and click on it"""
        box = await input_el.bounding_box()
        if box is None:
            return await input_el.click()

        x = box["x"] + int(box["width"] / random.uniform(1.1, 10))
        y = box["y"] + int(box["height"] / random.uniform(1.1, 10))

        await page.mouse.move(x, y, steps=random.randint(5, 30))
        await asyncio.sleep(random.uniform(0.2, 0.6))
        return await page.mouse.click(x, y)

    @abstractmethod
    async def _search(self, query, num_results=10):
        """Search web for links related to a query"""
        raise NotImplementedError


class PlaywrightSearchEngineLinkSearch(SearchEngineLinkSearch):
    """Abstract class to search the web using playwright and a search engine"""

    MAX_RESULTS_CONSIDERED_PER_PAGE = 10
    """Number of results considered per search engine page"""

    PAGE_LOAD_TIMEOUT = 60_000
    """Default page load timeout value in milliseconds"""

    _SC = StealthConfig(navigator_user_agent=False)

    def __init__(self, use_homepage=True, use_scrapling_stealth=False,
                 **launch_kwargs):
        """

        Parameters
        ----------
        use_homepage : bool, default=True
            If ``True``, the browser will be navigated to the search
            engine homepage and the query will be input into the search
            bar. If ``False``, the query will be embedded in the URL
            and the browser will navigate directly to the filled-out
            URL. By default, ``False``.
        use_scrapling_stealth : bool, default=False
            Option to use scrapling stealth scripts instead of
            tf-playwright-stealth. If set to ``True``, the `_SC` class
            attribute will be ignored. By default, ``False``.
        **launch_kwargs
            Keyword arguments to be passed to
            `playwright.chromium.launch`. For example, you can pass
            ``headless=False, slow_mo=50`` for a visualization of the
            search.
        """
        self.use_homepage = use_homepage
        self.use_scrapling_stealth = use_scrapling_stealth
        self.launch_kwargs = PWKwargs.launch_kwargs()
        self.launch_kwargs.update(launch_kwargs)
        self._browser = None

    async def _load_browser(self, pw_instance):
        """Launch a chromium instance and load a page"""
        self._browser = await pw_instance.chromium.launch(**self.launch_kwargs)

    async def _close_browser(self):
        """Close browser instance and reset internal attributes"""
        logger.trace("Closing browser...")
        if self._browser is None:
            return

        await self._browser.close()
        self._browser = None

    @asynccontextmanager
    async def _browser_page(self):
        """Get page to use for search"""
        page_kwargs = {"browser": self._browser, "stealth_config": self._SC,
                       "ignore_https_errors": True,  # no sensitive inputs
                       "timeout": self.PAGE_LOAD_TIMEOUT,
                       "use_scrapling_stealth": self.use_scrapling_stealth}

        async with pw_page(**page_kwargs) as page:
            yield page

    async def _search(self, query, num_results=10):
        """Search web for links related to a query"""
        logger.debug("Searching %s: %r", self._SE_NAME, query)
        num_results = min(num_results, self.MAX_RESULTS_CONSIDERED_PER_PAGE)

        async with self._browser_page() as page:
            if self.use_homepage:
                logger.trace("Navigating to %s homepage", self._SE_NAME)
                await _navigate_to_se_url(page, se_url=self._SE_URL,
                                          timeout=self.PAGE_LOAD_TIMEOUT)
                logger.trace("Performing %s search for query: %r",
                             self._SE_NAME, query)
                await self._perform_homepage_search(page, query)
            else:
                url = self._SE_QUERY_URL.format(quote(query))
                logger.trace("Submitting URL: %r", url)
                await _navigate_to_se_url(page, se_url=url,
                                          timeout=self.PAGE_LOAD_TIMEOUT)
            logger.trace("Extracting links for query: %r", query)
            return await self._extract_links(page, num_results, query)

    async def _get_links(self, queries, num_results):
        """Get links for multiple queries"""
        outer_task_name = asyncio.current_task().get_name()
        async with async_playwright() as pw_instance:
            await self._load_browser(pw_instance)
            searches = [
                asyncio.create_task(
                    self._skip_exc_search(query, num_results=num_results),
                    name=outer_task_name,
                )
                for query in queries
            ]
            logger.trace("Kicking off %s search for %d queries",
                         self._SE_NAME, len(searches))
            results = await asyncio.gather(*searches)
            logger.trace("Got results for link search:\n%r", results)
            await self._close_browser()
        return results

    async def _extract_links(self, page, num_results, query):
        """Extract links for top `num_results` on page"""
        await page.wait_for_load_state("networkidle",
                                       timeout=self.PAGE_LOAD_TIMEOUT)
        await page.wait_for_selector(self._SE_SR_TAG)
        locator = page.locator(self._SE_SR_TAG)
        count = await locator.count()
        links = []

        for i in range(count):
            element = locator.nth(i)
            try:
                link = await element.get_attribute("href")
                if link is not None:
                    links.append(link)
            except Exception:
                logger.exception("Skipped extracting link %d for query %r",
                                 i, query)

            if len(links) >= num_results:
                break

        return links

    @property
    @abstractmethod
    def _SE_URL(self):
        """str: Search engine URL"""
        raise NotImplementedError

    @property
    @abstractmethod
    def _SE_SR_TAG(self):
        """str: Search engine search results tag"""
        raise NotImplementedError

    @property
    @abstractmethod
    def _SE_QUERY_URL(self):
        """str: Search engine query URL template"""
        raise NotImplementedError

    @abstractmethod
    async def _perform_homepage_search(self, page, search_query):
        """Search query using search engine homepage"""
        raise NotImplementedError


class APISearchEngineLinkSearch(SearchEngineLinkSearch):
    """Abstract class to search the web using a search engine API"""

    API_KEY_VAR = None
    """Name of environment variable that should contain the API key"""

    def __init__(self, api_key=None):
        """

        Parameters
        ----------
        api_key : str, optional
            API key for search engine. If ``None``, will look up the API
            key using the :obj:`API_KEY_VAR` environment variable.
            By default, ``None``.
        """
        self.api_key = api_key or os.environ.get(self.API_KEY_VAR or "")


async def _navigate_to_se_url(page, se_url, timeout=90_000):
    """Navigate to search engine url"""
    await page.goto(se_url)
    logger.trace("Waiting for load")
    await page.wait_for_load_state("networkidle", timeout=timeout)



# -*- coding: utf-8 -*-
"""
ELM energy wizard
"""
from abc import ABC, abstractmethod
from time import perf_counter
import copy
import os
import json
import numpy as np

from elm.base import ApiBase
from elm.utilities.try_import import try_import


class EnergyWizardBase(ApiBase, ABC):
    """Base interface to ask OpenAI LLMs about energy research."""

    MODEL_ROLE = "You parse through articles to answer questions."
    """High level model role, somewhat redundant to MODEL_INSTRUCTION"""

    MODEL_INSTRUCTION = ('Use the information below to answer the subsequent '
                         'question. If the answer cannot be found in the '
                         'text, write "I could not find an answer."')
    """Prefix to the engineered prompt"""

    def __init__(self, model=None, token_budget=3500):
        """
        Parameters
        ----------
        model : str
            GPT model name, default is the DEFAULT_MODEL global var
        token_budget : int
            Number of tokens that can be embedded in the prompt. Note that the
            default budget for GPT-3.5-Turbo is 4096, but you want to subtract
            some tokens to account for the response budget.
        """

        super().__init__(model)
        self.token_budget = token_budget

    @abstractmethod
    def query_vector_db(self, query, limit=100):
        """Returns a list of strings and relatednesses, sorted from most
        related to least.

        Parameters
        ----------
        query : str
            Question being asked of GPT
        limit : int
            Number of top results to return.

        Returns
        -------
        strings : np.ndarray
            1D array of related strings
        score : np.ndarray
            1D array of float scores of strings
        idx : np.ndarray
            1D array of indices in the text corpus corresponding to the
            ranked strings/scores outputs.
        """

    def engineer_query(self,
                       query,
                       token_budget=None,
                       new_info_threshold=0.7,
                       convo=False
                       ):
        """Engineer a query for GPT using the corpus of information

        Parameters
        ----------
        query : str
            Question being asked of GPT
        token_budget : int
            Option to override the class init token budget.
        new_info_threshold : float
            New text added to the engineered query must contain at least this
            much new information. This helps prevent (for example) the table of
            contents being added multiple times.
        convo : bool
            Flag to perform semantic search with full conversation history
            (True) or just the single query (False). Call EnergyWizard.clear()
            to reset the chat history.

        Returns
        -------
        message : str
            Engineered question to GPT including information from corpus and
            the original query
        references : list
            The list of references (strs) used in the engineered prompt is
            returned here
        used_index : list
            Shows the indices of the documents used in making a query to the
            vector database
        vector_query_time : float
            measures vector database query time
        """

        self.messages.append({"role": "user", "content": query})

        if convo:
            # [1:] to not include the system role in the semantic search
            query = [f"{msg['role'].upper()}: {msg['content']}"
                     for msg in self.messages[1:]]

            query = '\n\n'.join(query)

        token_budget = token_budget or self.token_budget
        start_time = perf_counter()
        strings, _, idx = self.query_vector_db(query)
        end_time = perf_counter()
        vector_query_time = end_time - start_time
        message = copy.deepcopy(self.MODEL_INSTRUCTION)
        question = f"\n\nQuestion: {query}"
        used_index = []

        for string, i in zip(strings, idx):
            next_str = (f'\n\n"""\n{string}\n"""')
            token_usage = self.count_tokens(message + next_str + question,
                                            self.model)

            new_words = set(next_str.split(' '))
            additional_info = new_words - set(message.split(' '))
            new_info_frac = len(additional_info) / len(new_words)

            if new_info_frac > new_info_threshold:
                if token_usage > token_budget:
                    break
                else:
                    message += next_str
                    used_index.append(i)

        message = message + question
        used_index = np.array(used_index)
        references = self.make_ref_list(used_index)
        return message, references, used_index, vector_query_time

    @abstractmethod
    def make_ref_list(self, idx):
        """Make a reference list

        Parameters
        ----------
        used_index : np.ndarray
            Indices of the used text from the text corpus

        Returns
        -------
        ref_list : list
            A list of references (strs) used. Ideally, this is something like:
            ["{ref_title} ({ref_url})"]
        """

    def chat(self,
             query,
             stream=True,
             temperature=0,
             convo=False,
             token_budget=None,
             new_info_threshold=0.7,
             print_references=False,
             return_chat_obj=False
             ):
        """Answers a query by doing a semantic search of relevant text with
        embeddings and then sending engineered query to the LLM.

        Parameters
        ----------
        query : str
            Question being asked of EnergyWizard
        stream : bool
            Flag to print subsequent chunks of the response in a streaming
            fashion
        temperature : float
            GPT model temperature, a measure of response entropy from 0 to 1. 0
            is more reliable and nearly deterministic; 1 will give the model
            more creative freedom and may not return as factual of results.
        convo : bool
            Flag to perform semantic search with full conversation history
            (True) or just the single query (False). Call EnergyWizard.clear()
            to reset the chat history.
        token_budget : int
            Option to override the class init token budget.
        new_info_threshold : float
            New text added to the engineered query must contain at least this
            much new information. This helps prevent (for example) the table of
            contents being added multiple times.
        print_references : bool
            Flag to print references if EnergyWizard is initialized with a
            valid ``ref_col``.
        return_chat_obj : bool
            Flag to return the ChatCompletion object from OpenAI API instead of
            the message string.

        Returns
        -------
        response : str | ChatCompletion
            GPT response string ``if not return_chat_obj`` or OpenAI
            ChatCompletion object ``if return_chat_obj``
        query : str
            The engineered query asked of GPT including retrieved context
        references : list
            The list of references (strs) used in the engineered prompt is
            returned here
        performance : dict | None
            dictionary with keys of ``total_chat_time``,
            ``chat_completion_time`` and ``vectordb_query_time``. If
            return_chat_obj, this is None
        """
        start_chat_time = perf_counter()
        out = self.engineer_query(query, token_budget=token_budget,
                                  new_info_threshold=new_info_threshold,
                                  convo=convo)
        query, references, _, vector_query_time = out

        messages = [{"role": "system", "content": self.MODEL_ROLE},
                    {"role": "user", "content": query}]
        response_message = ''
        kwargs = dict(model=self.model,
                      messages=messages,
                      temperature=temperature,
                      stream=stream)

        start_completion_time = perf_counter()
        response = self._client.chat.completions.create(**kwargs)

        finish_completion_time = perf_counter()
        chat_completion_time = finish_completion_time - start_completion_time
        total_chat_time = finish_completion_time - start_chat_time

        performance = {
            "total_chat_time": total_chat_time,
            "chat_completion_time": chat_completion_time,
            "vectordb_query_time": vector_query_time
        }

        if return_chat_obj:
            return response, query, references, performance

        if stream:
            for chunk in response:
                chunk_msg = chunk.choices[0].delta.content or ""
                response_message += chunk_msg
                print(chunk_msg, end='')
        else:
            response_message = response.choices[0].message.content

        self.messages.append({'role': 'assistant',
                              'content': response_message})

        if any(references) and print_references:
            ref_msg = ('\n\nThe model was provided with the '
                       'following documents to support its answer:')
            ref_msg += '\n - ' + '\n - '.join(references)
            response_message += ref_msg
            if stream:
                print(ref_msg)

        end_time = perf_counter()
        total_chat_time = end_time - start_chat_time
        performance = {
            "total_chat_time": total_chat_time,
            "chat_completion_time": chat_completion_time,
            "vectordb_query_time": vector_query_time
        }

        return response_message, query, references, performance


class EnergyWizard(EnergyWizardBase):
    """Interface to ask OpenAI LLMs about energy research.

    This class is for execution on a local machine with a vector database in
    memory
    """

    def __init__(self, corpus, model=None, token_budget=3500, ref_col=None):
        """
        Parameters
        ----------
        corpus : pd.DataFrame
            Corpus of text in dataframe format. Must have columns "text" and
            "embedding".
        model : str
            GPT model name, default is the DEFAULT_MODEL global var
        token_budget : int
            Number of tokens that can be embedded in the prompt. Note that the
            default budget for GPT-3.5-Turbo is 4096, but you want to subtract
            some tokens to account for the response budget.
        ref_col : None | str
            Optional column label in the corpus that provides a reference text
            string for each chunk of text.
        """

        super().__init__(model, token_budget=token_budget)

        self.corpus = self.preflight_corpus(corpus)
        self.embedding_arr = np.vstack(self.corpus['embedding'].values)
        self.text_arr = self.corpus['text'].values
        self.ref_col = ref_col

    @staticmethod
    def preflight_corpus(corpus, required=('text', 'embedding')):
        """Run preflight checks on the text corpus.

        Parameters
        ----------
        corpus : pd.DataFrame
            Corpus of text in dataframe format. Must have columns "text" and
            "embedding".
        required : list | tuple
            Column names required to be in the corpus df

        Returns
        -------
        corpus : pd.DataFrame
            Corpus of text in dataframe format. Must have columns "text" and
            "embedding".
        """
        missing = [col for col in required if col not in corpus]
        if any(missing):
            msg = ('Text corpus must have {} columns but received '
                   'corpus with columns: {}'
                   .format(missing, list(corpus.columns)))
            raise KeyError(msg)

        if not isinstance(corpus.index.values[0], int):
            corpus['index'] = np.arange(len(corpus))
            corpus = corpus.set_index('index', drop=False)

        return corpus

    def cosine_dist(self, query_embedding):
        """Compute the cosine distance of the query embedding array vs. all of
        the embedding arrays of the full text corpus

        Parameters
        ----------
        query_embedding : np.ndarray
            1D array of the numerical embedding of the request query.

        Returns
        -------
        out : np.ndarray
            1D array with length equal to the number of entries in the text
            corpus. Each value is a distance score where smaller is closer
        """

        dot = np.dot(self.embedding_arr, query_embedding)
        norm1 = np.linalg.norm(query_embedding)
        norm2 = np.linalg.norm(self.embedding_arr, axis=1)

        out = 1 - (dot / (norm1 * norm2))

        return out

    def query_vector_db(self, query, limit=100):
        """Returns a list of strings and relatednesses, sorted from most
        related to least.

        Parameters
        ----------
        query : str
            Question being asked of GPT
        limit : int
            Number of top results to return.

        Returns
        -------
        strings : np.ndarray
            1D array of related strings
        score : np.ndarray
            1D array of float scores of strings
        idx : np.ndarray
            1D array of indices in the text corpus corresponding to the
            ranked strings/scores outputs.
        """

        embedding = self.get_embedding(query)
        scores = 1 - self.cosine_dist(embedding)
        best = np.argsort(scores)[::-1][:limit]

        strings = self.text_arr[best]
        scores = scores[best]

        return strings, scores, best

    def make_ref_list(self, idx):
        """Make a reference list

        Parameters
        ----------
        used_index : np.ndarray
            Indices of the used text from the text corpus

        Returns
        -------
        ref_list : list
            A list of references (strs) used. This takes information straight
            from ``ref_col``. Ideally, this is something like:
            ["{ref_title} ({ref_url})"]
        """
        ref_list = ''
        if self.ref_col is not None and self.ref_col in self.corpus:
            ref_list = list(self.corpus[self.ref_col].iloc[idx].unique())

        return ref_list


class EnergyWizardPostgres(EnergyWizardBase):
    """Interface to ask OpenAI LLMs about energy research.

    This class is for execution with a postgres vector database.
    Querying the database requires the use of the psycopg2 and
    boto3 python packages, environment variables ('EWIZ_DB_USER'
    and 'EWIZ_DB_PASSWORD') storing the db user and password, and
    the specification of other connection paremeters such as host,
    port, and name. The database has the following columns: id,
    embedding, chunks, and metadata.

    This class is designed as follows:
    Vector database: PostgreSQL database accessed using psycopg2.
    Query Embedding: AWS using boto3
    LLM Application: GPT4 via Azure deployment
    """
    EMBEDDING_MODEL = 'amazon.titan-embed-text-v1'

    TOKENIZER_ALIASES = {**EnergyWizardBase.TOKENIZER_ALIASES,
                         'ewiz-gpt-4': 'gpt-4'
                         }
    """Optional mappings for weird azure names to tiktoken/openai names."""

    DEFAULT_META_COLS = ['title', 'url', 'nrel_id', 'id']
    """Default columns to retrieve for metadata"""

    def __init__(self, db_host, db_port, db_name,
                 db_schema, db_table, probes=25,
                 meta_columns=None, cursor=None,
                 boto_client=None, model=None,
                 token_budget=3500, tag=False):
        """
        Parameters
        ----------
        db_host : str
            Host url for postgres database.
        db_port : str
            Port for postres database. ex: '5432'
        db_name : str
            Postgres database name.
        db_schema : str
            Schema name for postres database.
        db_table : str
            Table to query in Postgres database. Necessary columns: id,
            chunks, embedding, title, and url.
        probes : int
            Number of lists to search in vector database. Recommended
            value is sqrt(n_lists).
        meta_columns : list
            List of metadata columns to retrieve from database. Default
            query returns title, url, nrel_id, and id. nrel_id and id are
            necessary to correctly format references.
        cursor : psycopg2.extensions.cursor
            PostgreSQL database cursor used to execute queries.
        boto_client: botocore.client.BedrockRuntime
            AWS boto3 client used to access embedding model.
        model : str
            GPT model name, default is the DEFAULT_MODEL global var
        token_budget : int
            Number of tokens that can be embedded in the prompt. Note that the
            default budget for GPT-3.5-Turbo is 4096, but you want to subtract
            some tokens to account for the response budget.
        tag: bool
            Flag to add tag/metadata to text chunks before sending query to
            GPT.
        """
        boto3 = try_import('boto3')
        self.psycopg2 = try_import('psycopg2')

        if meta_columns is None:
            self.meta_columns = self.DEFAULT_META_COLS
        else:
            self.meta_columns = meta_columns

        assert 'id' in self.meta_columns, ("Please include the chunk id "
                                           "column: 'id'!")
        assert 'nrel_id' in self.meta_columns, ("Please include the document "
                                                "id column: 'nrel_id'!")

        if cursor is None:
            db_user = os.getenv("EWIZ_DB_USER")
            db_password = os.getenv('EWIZ_DB_PASSWORD')
            assert db_user is not None, "Must set EWIZ_DB_USER!"
            assert db_password is not None, "Must set EWIZ_DB_PASSWORD!"
            self.db_kwargs = dict(user=db_user, password=db_password,
                                  host=db_host, port=db_port,
                                  database=db_name)
            self.conn = self.psycopg2.connect(**self.db_kwargs)

            self.cursor = self.conn.cursor()
        else:
            self.cursor = cursor

        self.db_schema = db_schema
        self.db_table = db_table
        self.tag = tag
        self.probes = probes

        if boto_client is None:
            access_key = os.getenv('AWS_ACCESS_KEY_ID')
            secret = os.getenv('AWS_SECRET_ACCESS_KEY')
            session = os.getenv('AWS_SESSION_TOKEN')
            if access_key and secret and session:
                self._aws_client = boto3.client(service_name='bedrock-runtime',
                                                region_name='us-west-2',
                                                aws_access_key_id=access_key,
                                                aws_secret_access_key=secret,
                                                aws_session_token=session)
            else:
                self._aws_client = boto3.client(service_name='bedrock-runtime',
                                                region_name='us-west-2')
        else:
            self._aws_client = boto_client

        super().__init__(model, token_budget=token_budget)

    def get_embedding(self, text):
        """Get the 1D array (list) embedding of a text string
        as generated by specified AWS model.

        Parameters
        ----------
        text : str
            Text to embed

        Returns
        -------
        embedding : list
            List of float that represents the numerical embedding of the text
        """
        model_id = self.EMBEDDING_MODEL

        if 'cohere' in model_id:
            input_type = "search_query"

            body = json.dumps({"texts": [text],
                               "input_type": input_type})
        else:
            body = json.dumps({"inputText": text, })

        accept = 'application/json'
        content_type = 'application/json'

        response = self._aws_client.invoke_model(
            body=body,
            modelId=model_id,
            accept=accept,
            contentType=content_type
        )

        response_body = json.loads(response['body'].read())

        if 'cohere' in model_id:
            embedding = response_body.get('embeddings')[0]
        else:
            embedding = response_body.get('embedding')

        return embedding

    @staticmethod
    def _add_tag(meta):
        """Function to add tag/metadata to text strings before
        sending query to GPT.

        Parameters
        ----------
        meta : tuple
            Text values to include in tag (title, authors, year)

        Returns
        -------
        tag : str
            Text string containing provided metadata.
        """
        title, authors, year = meta
        if authors and year:
            tag = (f"Title: {title}\n"
                   f"Authors: {authors}\n"
                   f"Publication Year: {year}\n\n"
                   )
        else:
            tag = f"Title: {title}\n\n"

        return tag

    def query_vector_db(self, query, limit=100):
        """Returns a list of strings and relatednesses, sorted from most
        related to least. SQL query uses a context handler and rollback
        to ensure a failed query does not interupt future questions from
        the user. Ex: a user submitting a new question before the first
        one completes will close the cursor preventing future database
        access.

        Parameters
        ----------
        query : str
            Question being asked of GPT
        limit : int
            Number of top results to return.

        Returns
        -------
        strings : np.ndarray
            1D array of related strings
        score : np.ndarray
            1D array of float scores of strings
        ids : np.ndarray
            1D array of IDs in the text corpus corresponding to the
            ranked strings/scores outputs.
        """

        query_embedding = self.get_embedding(query)

        with self.psycopg2.connect(**self.db_kwargs) as conn:
            cursor = conn.cursor()
            try:
                cursor.execute(f"SET LOCAL ivfflat.probes = {self.probes};"
                               f"SELECT {self.db_table}.id, "
                               f"{self.db_table}.chunks, "
                               f"{self.db_table}.embedding "
                               "<=> %s::vector as score, "
                               f"{self.db_table}.title, "
                               f"{self.db_table}.authors, "
                               f"{self.db_table}.year "
                               f"FROM {self.db_schema}.{self.db_table} "
                               "ORDER BY embedding <=> %s::vector LIMIT %s;",
                               (query_embedding, query_embedding, limit,), )
            except Exception as exc:
                conn.rollback()
                msg = (f'Received error when querying the postgres '
                       f'vector database: {exc}')
                raise RuntimeError(msg) from exc
            else:
                conn.commit()
                result = cursor.fetchall()

        if self.tag:
            strings = [self._add_tag(s[3:]) + s[1] for s in result]
        else:
            strings = [s[1] for s in result]

        scores = [s[2] for s in result]
        best = [s[0] for s in result]

        return strings, scores, best

    def _format_refs(self, refs, ids):
        """Parse and nicely format a reference dictionary into a list of well
        formatted string representations

        Parameters
        ----------
        refs : list
            List of references returned from the vector db
        ids : np.ndarray
            IDs of the used text from the text corpus sorted by embedding
            relevance.

        Returns
        -------
        out : list
            Unique ordered list of references (most relevant first)
        """

        ref_list = []
        for item in refs:
            ref_dict = {col: str(value).replace(chr(34), '')
                        for col, value in zip(self.meta_columns, item)}

            ref_list.append(ref_dict)

        assert len(ref_list) > 0, ("The Wizard did not return any "
                                   "references. Please check your database "
                                   "connection or query.")

        unique_ref_list = []
        unique_nrel_ids = set()
        for ref_dict in ref_list:
            if ref_dict['nrel_id'] in unique_nrel_ids:
                continue
            unique_ref_list.append(ref_dict)
            unique_nrel_ids.add(ref_dict['nrel_id'])

        ref_list = unique_ref_list

        if 'id' in ref_list[0]:
            ids_list = list(ids)
            sorted_ref_list = []
            for ref_id in ids_list:
                for ref_dict in ref_list:
                    if ref_dict['id'] == ref_id:
                        sorted_ref_list.append(ref_dict)
                        break
            ref_list = sorted_ref_list

        ref_list = [json.dumps(ref) for ref in ref_list]

        return ref_list

    def make_ref_list(self, ids):
        """Make a reference list. SQL query uses a context handler and
        rollback to ensure a failed query does not interupt future questions
        from the user. Ex: a user submitting a new question before the first
        one completes will close the cursor preventing future database
        access.

        Parameters
        ----------
        ids : np.ndarray
            IDs of the used text from the text corpus

        Returns
        -------
        ref_list : list
            A list of references (strs) used. Ideally, this is something like:
            ["{ref_title} ({ref_url})"]
        """

        placeholders = ', '.join(['%s'] * len(ids))
        columns_str = ', '.join([f"{self.db_table}.{c}"
                                 for c in self.meta_columns])

        sql_query = (f"SELECT {columns_str} "
                     f"FROM {self.db_schema}.{self.db_table} "
                     f"WHERE {self.db_table}.id IN (" + placeholders + ")")

        with self.psycopg2.connect(**self.db_kwargs) as conn:
            cursor = conn.cursor()
            try:
                cursor.execute(sql_query, ids)
            except Exception as exc:
                conn.rollback()
                msg = (f'Received error when querying the postgres '
                       f'vector database: {exc}')
                raise RuntimeError(msg) from exc
            else:
                conn.commit()
                refs = cursor.fetchall()

        ref_list = self._format_refs(refs, ids)

        return ref_list



# -*- coding: utf-8 -*-
"""
ELM text embedding
"""
import openai
import re
import os
import logging

from elm.base import ApiBase
from elm.chunk import Chunker


logger = logging.getLogger(__name__)


class ChunkAndEmbed(ApiBase):
    """Class to chunk text data and create embeddings"""

    DEFAULT_MODEL = 'text-embedding-ada-002'
    """Default model to do embeddings."""

    def __init__(self, text, model=None, **chunk_kwargs):
        """
        Parameters
        ----------
        text : str
            Single continuous piece of text to chunk up by paragraph and embed
            or filepath to .txt file containing one piece of text.
        model : None | str
            Optional specification of OpenAI model to use. Default is
            cls.DEFAULT_MODEL
        chunk_kwargs : dict | None
            kwargs for initialization of :class:`elm.chunk.Chunker`
        """

        super().__init__(model)

        self.text = text

        if os.path.isfile(text):
            logger.info('Loading text file: {}'.format(text))
            with open(text, 'r') as f:
                self.text = f.read()

        assert isinstance(self.text, str)
        self.text = self.clean_tables(self.text)

        self.text_chunks = Chunker(self.text, **chunk_kwargs)

    @staticmethod
    def clean_tables(text):
        """Make sure that table headers are in the same paragraph as the table
        itself. Typically, tables are looked for with pipes and hyphens, which
        is how GPT cleans tables in text."""

        # looks for "Table N.", should expand to other formats with additional
        # regex patterns later
        table_regex = r"^Table [0-9]+."

        lines = text.split('\n')
        for idx, line in enumerate(lines[:-2]):
            next_line_1 = lines[idx + 1]
            next_line_2 = lines[idx + 2]
            match = re.search(table_regex, line)
            cond1 = match is not None
            cond2 = next_line_1.strip() == ''
            cond3 = next_line_2.startswith('|')

            if all([cond1, cond2, cond3]):
                lines[idx + 1] = line
                lines[idx] = ''

        return '\n'.join(lines)

    def run(self, rate_limit=175e3):
        """Run text embedding in serial

        Parameters
        ----------
        rate_limit : float
            OpenAI API rate limit (tokens / minute). Note that the
            embedding limit is 350k as of 4/2023, but we're using a large
            factor of safety (~1/2) because we can only count the tokens on the
            input side and assume the output is about the same count.

        Returns
        -------
        embedding : list
            List of 1D arrays representing the embeddings for all text chunks
        """

        logger.info('Embedding {} text chunks...'
                    .format(len(self.text_chunks)))

        embeddings = []
        for i, chunk in enumerate(self.text_chunks):
            req = {"input": chunk, "model": self.model}

            if 'azure' in str(openai.api_type).lower():
                req['engine'] = self.model

            out = self.call_api(self.EMBEDDING_URL, self.HEADERS, req)

            try:
                out = out['data'][0]['embedding']
                embeddings.append(out)
            except Exception:
                msg = ('Could not get embeddings for chunk {}, '
                       'received API response: {}'.format(i + 1, out))
                logger.error(msg)
                embeddings.append(None)

        logger.info('Finished all embeddings.')

        return embeddings

    async def run_async(self, rate_limit=175e3):
        """Run text embedding on chunks asynchronously

        NOTE: you need to call this using the await command in ipython or
        jupyter, e.g.: `out = await ChunkAndEmbed.run_async()`

        Parameters
        ----------
        rate_limit : float
            OpenAI API rate limit (tokens / minute). Note that the
            embedding limit is 350k as of 4/2023, but we're using a large
            factor of safety (~1/2) because we can only count the tokens on the
            input side and assume the output is about the same count.

        Returns
        -------
        embedding : list
            List of 1D arrays representing the embeddings for all text chunks
        """

        logger.info('Embedding {} text chunks...'
                    .format(len(self.text_chunks)))

        all_request_jsons = []
        for chunk in self.text_chunks:
            req = {"input": chunk, "model": self.model}

            if 'azure' in str(openai.api_type).lower():
                req['engine'] = self.model

            all_request_jsons.append(req)

        embeddings = await self.call_api_async(self.EMBEDDING_URL,
                                               self.HEADERS,
                                               all_request_jsons,
                                               rate_limit=rate_limit)

        for i, chunk in enumerate(embeddings):
            try:
                embeddings[i] = chunk['data'][0]['embedding']
            except Exception:
                msg = ('Could not get embeddings for chunk {}, '
                       'received API response: {}'.format(i + 1, chunk))
                logger.error(msg)
                embeddings[i] = None

        logger.info('Finished all embeddings.')

        return embeddings



# -*- coding: utf-8 -*-
"""ELM Web HTML loading with Playwright

We use Playwright so that javascript text is rendered before we scrape.
"""
import logging
from contextlib import AsyncExitStack

from rebrowser_playwright.async_api import async_playwright
from rebrowser_playwright.async_api import Error as PlaywrightError
from rebrowser_playwright.async_api import (
    TimeoutError as PlaywrightTimeoutError
)

from elm.web.utilities import pw_page, PWKwargs


logger = logging.getLogger(__name__)


async def load_html_with_pw(url, browser_semaphore=None, # pragma: no cover
                            timeout=90_000, use_scrapling_stealth=False,
                            **pw_launch_kwargs):
    """Extract HTML from URL using Playwright.

    Parameters
    ----------
    url : str
        URL to pull HTML for.
    browser_semaphore : asyncio.Semaphore, optional
        Semaphore instance that can be used to limit the number of
        playwright browsers open concurrently. If ``None``, no limits
        are applied. By default, ``None``.
    timeout : int, optional
        Maximum time to wait for page loading state time in
        milliseconds. Pass `0` to disable timeout.
        By default, ``90,000``.
    use_scrapling_stealth : bool, default=False
        Option to use scrapling stealth scripts instead of
        tf-playwright-stealth. By default, ``False``.
    **pw_launch_kwargs
        Keyword-value argument pairs to pass to
        :meth:`async_playwright.chromium.launch`.

    Returns
    -------
    str
        HTML from page.
    """
    try:
        text = await _load_html(url, browser_semaphore=browser_semaphore,
                                timeout=timeout,
                                use_scrapling_stealth=use_scrapling_stealth,
                                **pw_launch_kwargs)
    except (PlaywrightError, PlaywrightTimeoutError):
        text = ""
    return text


async def _load_html( url, browser_semaphore=None, # pragma: no cover
                     timeout=90_000, use_scrapling_stealth=False,
                     **pw_launch_kwargs):
    """Load html using playwright"""
    logger.trace("`_load_html` pw_launch_kwargs=%r", pw_launch_kwargs)
    logger.trace("browser_semaphore=%r", browser_semaphore)

    if browser_semaphore is None:
        browser_semaphore = AsyncExitStack()

    launch_kwargs = PWKwargs.launch_kwargs()
    launch_kwargs.update(pw_launch_kwargs)

    logger.trace("Loading HTML using playwright")
    async with async_playwright() as p, browser_semaphore:
        logger.trace("launching chromium; browser_semaphore=%r",
                     browser_semaphore)
        browser = await p.chromium.launch(**launch_kwargs)
        page_kwargs = {"browser": browser, "intercept_routes": True,
                       "timeout": timeout, "ignore_https_errors": True,
                       "use_scrapling_stealth": use_scrapling_stealth}
        async with pw_page(**page_kwargs) as page:
            logger.trace("Navigating to: %r", url)
            await page.goto(url)
            logger.trace("Waiting for load with timeout: %r", timeout)
            await page.wait_for_load_state("networkidle", timeout=timeout)
            text = await page.content()

    return text



# -*- coding: utf-8 -*-
"""ELM Web Scraping - Tavily API search"""
import logging

from tavily import TavilyClient

from elm.web.search.base import APISearchEngineLinkSearch


logger = logging.getLogger(__name__)


class APITavilySearch(APISearchEngineLinkSearch):
    """Search the web for links using the Tavily API"""

    _SE_NAME = "Tavily API"

    API_KEY_VAR = "TAVILY_API_KEY"
    """Environment variable that should contain the Tavily API key"""

    async def _search(self, query, num_results=10):
        """Search web for links related to a query"""

        client = TavilyClient(api_key=self.api_key)
        response = client.search(query=query, max_results=num_results)
        results = response.get("results", [])
        return list(filter(None, (info.get('url', "").replace("+", "%20")
                                  for info in results)))



# -*- coding: utf-8 -*-
"""ELM Web Scraping - Yahoo search"""
import random
import asyncio
import logging

from elm.web.search.base import PlaywrightSearchEngineLinkSearch


logger = logging.getLogger(__name__)


class PlaywrightYahooLinkSearch(PlaywrightSearchEngineLinkSearch):
    """Search for top links on the main Yahoo search engine"""

    MAX_RESULTS_CONSIDERED_PER_PAGE = 10
    """Number of results considered per Yahoo page.

    This value used to be 10, but the addition of extra divs like a set
    of youtube links has brought this number down."""

    _SE_NAME = "Yahoo"
    _SE_URL = "https://search.yahoo.com/"
    _SE_SR_TAG = '[referrerpolicy="origin"]'
    _SE_QUERY_URL = (
        "https://search.yahoo.com/search?p={}&fr=sfp&fr2=p%3As%2Cv%3Asfp"
    )

    async def _perform_homepage_search(self, page, search_query):
        """Fill in search bar with user query and hit enter"""
        await self._move_mouse(page)

        logger.trace("Clicking on search bar")
        search_bar = page.locator('[id="yschsp"]')
        await self._move_and_click(page, search_bar)
        await asyncio.sleep(random.uniform(0.5, 1.5))

        logger.trace("Typing in query: %r", search_query)
        await page.keyboard.type(search_query, delay=random.randint(80, 150))
        await asyncio.sleep(random.uniform(0.5, 1.5))

        logger.trace("Hitting enter for query: %r", search_query)
        await page.keyboard.press('Enter')



# -*- coding: utf-8 -*-
"""ELM Ordinance ngram text validation

This check helps validate that the LLM extracted text from the original
document and did not make it up itself.
"""
import nltk
from nltk.tokenize import word_tokenize
from nltk.tokenize import sent_tokenize
from nltk.corpus import stopwords
from nltk.util import ngrams


nltk.download("punkt", quiet=True)
nltk.download("punkt_tab", quiet=True)
nltk.download("stopwords", quiet=True)
STOP_WORDS = set(stopwords.words("english"))
PUNCTUATIONS = {'"', ".", "(", ")", ",", "?", ";", ":", "''", "``"}


def _check_word(word):
    """``True`` if a word is not a stop word or a punctuation."""
    return word not in STOP_WORDS and word not in PUNCTUATIONS


def _filtered_words(sentence):
    """Filter out common words and punctuations."""
    return [
        word.casefold()
        for word in word_tokenize(sentence)
        if _check_word(word.casefold())
    ]


def convert_text_to_sentence_ngrams(text, n):
    """Convert input text to a list of ngrams.

    The text is first split byu sentence, after which each sentence is
    converted into ngrams. The ngrams for all sentences are combined and
    returned.

    Parameters
    ----------
    text : str
        Input text containing one or more sentences.
    n : int
        Number of words to include per ngram.

    Returns
    -------
    list
        List of tuples, where each tuple is an ngram from the original
        text.
    """
    all_ngrams = []
    sentences = sent_tokenize(text)
    for sentence in sentences:
        words = _filtered_words(sentence)
        all_ngrams += list(ngrams(words, n))
    return all_ngrams


def sentence_ngram_containment(original, test, n):
    """Fraction of sentence ngrams from the test text found in the original.

    Parameters
    ----------
    original : str
        Original (superset) text. Ngrams from the `test` text will be
        checked against this text.
    test : str
        Test (sub) text. Ngrams from this text will be searched for in
        the original text, and the fraction of these ngrams that are
        found in the original text will be returned.
    n : int
        Number of words to include per ngram.

    Returns
    -------
    float
        Fraction of ngrams from the `test` input that were found in the
        `original` text. Always returns ``True`` if test has no ngrams.
    """
    ngrams_test = convert_text_to_sentence_ngrams(test, n)
    num_test_ngrams = len(ngrams_test)
    if not num_test_ngrams:
        return True

    ngrams_original = set(convert_text_to_sentence_ngrams(original, n))
    num_ngrams_found = sum(t in ngrams_original for t in ngrams_test)
    return num_ngrams_found / num_test_ngrams



# -*- coding: utf-8 -*-
"""ELM Ordinances OpenAI service amd utils."""
import logging

import openai

from elm.base import ApiBase
from elm.ords.services.base import RateLimitedService
from elm.ords.services.usage import TimeBoundedUsageTracker
from elm.utilities.retry import async_retry_with_exponential_backoff


logger = logging.getLogger(__name__)


def usage_from_response(current_usage, response):
    """OpenAI usage parser.

    Parameters
    ----------
    current_usage : dict
        Dictionary containing current usage information. For OpenAI
        trackers, this may contain the keys ``"requests"``,
        ``"prompt_tokens"``, and ``"response_tokens"`` if there is
        already existing tracking information. Empty dictionaries are
        allowed, in which case the three keys above will be added to
        this input.
    response : openai.Completion
        OpenAI Completion object. Must contain a ``usage`` attribute
        that

    Returns
    -------
    dict
        Dictionary with updated usage statistics.
    """
    current_usage["requests"] = current_usage.get("requests", 0) + 1
    current_usage["prompt_tokens"] = (
        current_usage.get("prompt_tokens", 0) + response.usage.prompt_tokens
    )
    current_usage["response_tokens"] = (
        current_usage.get("response_tokens", 0)
        + response.usage.completion_tokens
    )
    return current_usage


def count_tokens(messages, model):
    """Count the number of tokens in an outgoing set of messages.

    Parameters
    ----------
    messages : list
        A list of message objects, where the latter is represented
        using a dictionary. Each message dictionary must have a
        "content" key containing the string to count tokens for.
    model : str
        The OpenAI model being used. This input will be passed to
        :func:`tiktoken.encoding_for_model`.

    Returns
    -------
    int
        Total number of tokens in the set of messages outgoing to
        OpenAI.

    References
    ----------
    https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb
    """
    message_total = sum(
        ApiBase.count_tokens(message["content"], model=model) + 4
        for message in messages
    )
    return message_total + 3


class OpenAIService(RateLimitedService):
    """OpenAI Chat GPT query service

    Purpose:
        Orchestrate OpenAI API calls.
    Responsibilities:
        1. Monitor OpenAI call queue.
        2. Submit calls to OpenAI API if rate limit has not been
           exceeded.
        3. Track token usage, both instantaneous (rate) and total (if
           user requests it).
        4. Parse responses into `str` and pass back to calling function.
    Key Relationships:
        Must be activated with
        :class:`~elm.ords.services.provider.RunningAsyncServices`
        context.

    .. end desc
    """

    def __init__(self, client, rate_limit=1e3, rate_tracker=None):
        """

        Parameters
        ----------
        client : openai.AsyncOpenAI | openai.AsyncAzureOpenAI
            Async OpenAI client instance. Must have an async
            `client.chat.completions.create` method.
        rate_limit : int | float, optional
            Token rate limit (typically per minute, but the time
            interval is ultimately controlled by the `rate_tracker`
            instance). By default, ``1e3``.
        rate_tracker : TimeBoundedUsageTracker, optional
            A TimeBoundedUsageTracker instance. This will be used to
            track usage per time interval and compare to `rate_limit`.
            If ``None``, a `TimeBoundedUsageTracker` instance is created
            with default parameters. By default, ``None``.
        """
        super().__init__(rate_limit, rate_tracker or TimeBoundedUsageTracker())
        self.client = client

    async def process(
        self, usage_tracker=None, usage_sub_label="default", *, model, **kwargs
    ):
        """Process a call to OpenAI Chat GPT.

        Note that this method automatically retries queries (with
        backoff) if a rate limit error is throw by the API.

        Parameters
        ----------
        model : str
            OpenAI GPT model to query.
        usage_tracker : `elm.ords.services.usage.UsageTracker`, optional
            UsageTracker instance. Providing this input will update your
            tracker with this call's token usage info.
            By default, ``None``.
        usage_sub_label : str, optional
            Optional label to categorize usage under. This can be used
            to track usage related to certain categories.
            By default, ``"default"``.
        **kwargs
            Keyword arguments to be passed to
            `client.chat.completions.create`.

        Returns
        -------
        str | None
            Chat GPT response as a string, or ``None`` if the call
            failed.
        """
        self._record_prompt_tokens(model, kwargs)
        response = await self._call_gpt(model=model, **kwargs)
        self._record_completion_tokens(response)
        self._record_usage(response, usage_tracker, usage_sub_label)
        return _get_response_message(response)

    def _record_prompt_tokens(self, model, kwargs):
        """Add prompt token count to rate tracker"""
        num_tokens = count_tokens(kwargs.get("messages", []), model)
        self.rate_tracker.add(num_tokens)

    def _record_usage(self, response, usage_tracker, usage_sub_label):
        """Record token usage for user"""
        if usage_tracker is None:
            return
        usage_tracker.update_from_model(response, sub_label=usage_sub_label)

    def _record_completion_tokens(self, response):
        """Add completion token count to rate tracker"""
        if response is None:
            return
        self.rate_tracker.add(response.usage.completion_tokens)

    @async_retry_with_exponential_backoff()
    async def _call_gpt(self, **kwargs):
        """Query Chat GPT with user inputs"""
        try:
            return await self.client.chat.completions.create(**kwargs)
        except openai.BadRequestError as e:
            logger.error("Got 'BadRequestError':")
            logger.exception(e)


def _get_response_message(response):
    """Get message as string from response object"""
    if response is None:
        return None
    return response.choices[0].message.content



# -*- coding: utf-8 -*-
"""Custom Exceptions and Errors for ELM Ordinances. """
import logging

from elm.exceptions import ELMError


logger = logging.getLogger("elm")


class ELMOrdsError(ELMError):
    """Generic ELM Ordinance Error."""

    def __init__(self, *args, **kwargs):
        """Init exception and broadcast message to logger."""
        super().__init__(*args, **kwargs)
        if args:
            logger.error(str(args[0]), stacklevel=2)


class ELMOrdsNotInitializedError(ELMOrdsError):
    """ELM Ordinances not initialized error."""


class ELMOrdsValueError(ELMOrdsError, ValueError):
    """ELM Ordinances ValueError."""


class ELMOrdsRuntimeError(ELMOrdsError, RuntimeError):
    """ELM Ordinances RuntimeError."""



# -*- coding: utf-8 -*-
"""ELM Ordinance Decision Tree Graph setup functions."""
import networkx as nx


_SECTION_PROMPT = (
    'The value of the "section" key should be a string representing the '
    "title of the section (including numerical labels), if it's given, "
    "and `null` otherwise."
)
_COMMENT_PROMPT = (
    'The value of the "comment" key should be a one-sentence explanation '
    "of how you determined the value, if you think it is necessary "
    "(`null` otherwise)."
)
EXTRACT_ORIGINAL_TEXT_PROMPT = (
    "Can you extract the raw text with original formatting "
    "that states how close I can site {wes_type} to {feature}? "
)


def _setup_graph_no_nodes(**kwargs):
    return nx.DiGraph(
        SECTION_PROMPT=_SECTION_PROMPT,
        COMMENT_PROMPT=_COMMENT_PROMPT,
        **kwargs
    )


def llm_response_starts_with_yes(response):
    """Check if LLM response begins with "yes" (case-insensitive)

    Parameters
    ----------
    response : str
        LLM response string.

    Returns
    -------
    bool
        `True` if LLM response begins with "Yes".
    """
    return response.lower().startswith("yes")


def llm_response_starts_with_no(response):
    """Check if LLM response begins with "no" (case-insensitive)

    Parameters
    ----------
    response : str
        LLM response string.

    Returns
    -------
    bool
        `True` if LLM response begins with "No".
    """
    return response.lower().startswith("no")


def llm_response_does_not_start_with_no(response):
    """Check if LLM response does not start with "no" (case-insensitive)

    Parameters
    ----------
    response : str
        LLM response string.

    Returns
    -------
    bool
        `True` if LLM response does not begin with "No".
    """
    return not llm_response_starts_with_no(response)


def setup_graph_wes_types(**kwargs):
    """Setup Graph to get the largest turbine size in the ordinance text.

    Parameters
    ----------
    **kwargs
        Keyword-value pairs to add to graph.

    Returns
    -------
    nx.DiGraph
        Graph instance that can be used to initialize an
        `elm.tree.DecisionTree`.
    """
    G = _setup_graph_no_nodes(**kwargs)

    G.add_node(
        "init",
        prompt=(
            "Does the following text distinguish between multiple "
            "turbine sizes? Distinctions are often made as 'small' vs 'large' "
            "wind energy conversion systems or actual MW values. "
            "Begin your response with either 'Yes' or 'No' and explain your "
            "answer."
            '\n\n"""\n{text}\n"""'
        ),
    )

    G.add_edge("init", "get_text", condition=llm_response_starts_with_yes)
    G.add_node(
        "get_text",
        prompt=(
            "What are the different turbine sizes this text mentions? "
            "List them in order of increasing size."
        ),
    )
    G.add_edge("get_text", "final")
    G.add_node(
        "final",
        prompt=(
            "Respond based on our entire conversation so far. Return your "
            "answer in JSON format (not markdown). Your JSON file must "
            'include exactly two keys. The keys are "largest_wes_type" and '
            '"explanation". The value of the "largest_wes_type" key should '
            "be a string that labels the largest wind energy conversion "
            'system mentioned in the text. The value of the "explanation" '
            "key should be a string containing a short explanation for your "
            "choice."
        ),
    )
    return G


def setup_base_graph(**kwargs):
    """Setup Graph to get setback ordinance text for a particular feature.

    Parameters
    ----------
    **kwargs
        Keyword-value pairs to add to graph.

    Returns
    -------
    nx.DiGraph
        Graph instance that can be used to initialize an
        `elm.tree.DecisionTree`.
    """
    G = _setup_graph_no_nodes(**kwargs)

    G.add_node(
        "init",
        prompt=(
            "Is there text in the following legal document that describes "
            "how close I can site or how far I have to setback "
            "{wes_type} to {feature}? {feature_clarifications}"
            "Pay extra attention to clarifying text found in parentheses "
            "and footnotes. Begin your response with either 'Yes' or 'No' "
            "and explain your answer."
            '\n\n"""\n{text}\n"""'
        ),
    )

    G.add_edge(
        "init", "get_text", condition=llm_response_does_not_start_with_no
    )
    G.add_node("get_text", prompt=EXTRACT_ORIGINAL_TEXT_PROMPT)

    return G


def setup_participating_owner(**kwargs):
    """Setup Graph to check for participating vs non-participating owner
    setbacks for a feature.

    Parameters
    ----------
    **kwargs
        Keyword-value pairs to add to graph.

    Returns
    -------
    nx.DiGraph
        Graph instance that can be used to initialize an
        `elm.tree.DecisionTree`.
    """
    G = _setup_graph_no_nodes(**kwargs)

    G.add_node(
        "init",
        prompt=(
            "Does the ordinance for {feature} setbacks explicitly specify "
            "a value that applies to participating owners? Occupying owners "
            "are not participating owners unless explicitly mentioned in the "
            "text. Justify your answer by quoting the raw text directly."
        ),
    )
    G.add_edge("init", "non_part")
    G.add_node(
        "non_part",
        prompt=(
            "Does the ordinance for {feature} setbacks explicitly specify "
            "a value that applies to non-participating owners? Non-occupying "
            "owners are not non-participating owners unless explicitly "
            "mentioned in the text. Justify your answer by quoting the raw "
            "text directly."
        ),
    )
    G.add_edge("non_part", "final")
    G.add_node(
        "final",
        prompt=(
            "Now we are ready to extract structured data. Respond based on "
            "our entire conversation so far. Return your answer in JSON "
            "format (not markdown). Your JSON file must include exactly two "
            'keys. The keys are "participating" and "non-participating". The '
            'value of the "participating" key should be a string containing '
            "the raw text with original formatting from the ordinance that "
            "applies to participating owners or `null` if there was no such "
            'text. The value of the "non-participating" key should be a '
            "string containing the raw text with original formatting from the "
            "ordinance that applies to non-participating owners or simply the "
            "full ordinance if the text did not make the distinction between "
            "participating and non-participating owners."
        ),
    )
    return G


def setup_multiplier(**kwargs):
    """Setup Graph to extract a setbacks multiplier values for a feature.

    Parameters
    ----------
    **kwargs
        Keyword-value pairs to add to graph.

    Returns
    -------
    nx.DiGraph
        Graph instance that can be used to initialize an
        `elm.tree.DecisionTree`.
    """
    G = _setup_graph_no_nodes(**kwargs)

    G.add_node(
        "init",
        prompt=(
            "We will attempt to extract structured data for this ordinance. "
            "Let's think step by step. Does the text mention a multiplier "
            "that should be applied to a turbine dimension (e.g. height, "
            "rotor diameter, etc) to compute the setback distance from "
            "{feature}? Ignore any text related to {ignore_features}. "
            "Remember that 1 is a valid multiplier, and treat any mention of "
            "'fall zone' as a system height multiplier of 1. Begin your "
            "response with either 'Yes' or 'No' and explain your answer."
        ),
    )
    G.add_edge("init", "no_multiplier", condition=llm_response_starts_with_no)
    G.add_node(
        "no_multiplier",
        prompt=(
            "Does the ordinance give the setback from {feature} as a fixed "
            "distance value? Explain yourself."
        ),
    )
    G.add_edge("no_multiplier", "out_static")
    G.add_node(
        "out_static",
        prompt=(
            "Now we are ready to extract structured data. Respond based on "
            "our entire conversation so far. Return your answer in JSON "
            "format (not markdown). Your JSON file must include exactly "
            'four keys. The keys are "fixed_value", "units", "section", '
            '"comment". The value of the "fixed_value" key should be a '
            "numerical value corresponding to the setback distance value "
            "from {feature} or `null` if there was no such value. The value "
            'of the "units" key should be a string corresponding to the units '
            "of the setback distance value from {feature} or `null` if there "
            'was no such value. {SECTION_PROMPT} The value of the "comment" '
            "key should be a one-sentence explanation of how you determined "
            "the value, or a short description of the ordinance itself if no "
            "multiplier or static setback value was found."
        ),
    )
    G.add_edge("init", "mult_single", condition=llm_response_starts_with_yes)

    G.add_node(
        "mult_single",
        prompt=(
            "Are multiple values given for the multiplier used to "
            "compute the setback distance value from {feature}? If so, "
            "select and state the largest one. Otherwise, repeat the single "
            "multiplier value that was given in the text. "
        ),
    )
    G.add_edge("mult_single", "mult_type")
    G.add_node(
        "mult_type",
        prompt=(
            "What should the multiplier be applied to? Common acronyms "
            "include RD for rotor diameter and HH for hub height. Remember "
            "that system/total height is the tip-hight of the turbine. "
            "Select a value from the following list and explain yourself: "
            "['tip-height-multiplier', 'hub-height-multiplier', "
            "'rotor-diameter-multiplier]"
        ),
    )

    G.add_edge("mult_type", "adder")
    G.add_node(
        "adder",
        prompt=(
            "Does the ordinance include a static distance value that "
            "should be added to the result of the multiplication? Do not "
            "confuse this value with static setback requirements. Ignore text "
            "with clauses such as 'no lesser than', 'no greater than', "
            "'the lesser of', or 'the greater of'. Begin your response with "
            "either 'Yes' or 'No' and explain your answer, stating the adder "
            "value if it exists."
        ),
    )
    G.add_edge("adder", "out_mult", condition=llm_response_starts_with_no)
    G.add_edge("adder", "adder_eq", condition=llm_response_starts_with_yes)

    G.add_node(
        "adder_eq",
        prompt=(
            "We are only interested in adders that satisfy the following "
            "equation: 'multiplier * turbine_dimension + <adder>'. Does the "
            "adder value you identified satisfy this equation? Begin your "
            "response with either 'Yes' or 'No' and explain your answer."
        ),
    )
    G.add_edge("adder_eq", "out_mult", condition=llm_response_starts_with_no)
    G.add_edge(
        "adder_eq",
        "conversion",
        condition=llm_response_starts_with_yes,
    )
    G.add_node(
        "conversion",
        prompt=(
            "If the adder value is not given in feet, convert "
            "it to feet (remember that there are 3.28084 feet in one meter "
            "and 5280 feet in one mile). Show your work step-by-step "
            "if you had to perform a conversion."
        ),
    )
    G.add_edge("conversion", "out_mult")

    G.add_node(
        "out_mult",
        prompt=(
            "Now we are ready to extract structured data. Respond based on "
            "our entire conversation so far. Return your answer in JSON "
            "format (not markdown). Your JSON file must include exactly five "
            'keys. The keys are "mult_value", "mult_type", "adder", '
            '"section", "comment". The value of the "mult_value" key should '
            "be a numerical value corresponding to the multiplier value we "
            'determined earlier. The value of the "mult_type" key should be '
            "a string corresponding to the dimension that the multiplier "
            "should be applied to, as we determined earlier. The value of "
            'the "adder" key should be a numerical value corresponding to '
            "the static value to be added to the total setback distance after "
            "multiplication, as we determined earlier, or `null` if there is "
            "no such value. {SECTION_PROMPT} {COMMENT_PROMPT}"
        ),
    )

    return G


def setup_conditional(**kwargs):
    """Setup Graph to extract min/max setback values (after mult) for a
    feature. These are typically given within the context of
    'the greater of' or 'the lesser of' clauses.

    Parameters
    ----------
    **kwargs
        Keyword-value pairs to add to graph.

    Returns
    -------
    nx.DiGraph
        Graph instance that can be used to initialize an
        `elm.tree.DecisionTree`.
    """
    G = _setup_graph_no_nodes(**kwargs)

    G.add_node(
        "init",
        prompt=(
            "We will attempt to extract structured data for this ordinance. "
            "Let's think step by step. Does the setback from {feature} "
            "mention a minimum or maximum static setback distance regardless "
            "of the outcome of the multiplier calculation? This is often "
            "phrased as 'the greater of' or 'the lesser of'. Do not confuse "
            "this value with static values to be added to multiplicative "
            "setbacks. Begin your response with either 'Yes' or 'No' and "
            "explain your answer."
        ),
    )

    G.add_edge("init", "conversions", condition=llm_response_starts_with_yes)
    G.add_node(
        "conversions",
        prompt=(
            "Tell me the minimum and/or maximum setback distances, "
            "converting to feet if necessary (remember that there are "
            "3.28084 feet in one meter and 5280 feet in one mile). "
            "Explain your answer and show your work if you had to perform "
            "a conversion."
        ),
    )

    G.add_edge("conversions", "out_condition")
    G.add_node(
        "out_condition",
        prompt=(
            "Now we are ready to extract structured data. Respond based "
            "on our entire conversation so far. Return your answer in JSON "
            "format (not markdown). Your JSON file must include exactly two "
            'keys. The keys are "min_dist" and "max_dist". The value of the '
            '"min_dist" key should be a numerical value corresponding to the '
            "minimum setback value from {feature} we determined earlier, or "
            '`null` if no such value exists. The value of the "max_dist" key '
            "should be a numerical value corresponding to the maximum setback "
            "value from {feature} we determined earlier, or `null` if no such "
            "value exists."
        ),
    )

    return G


def setup_graph_extra_restriction(**kwargs):
    """Setup Graph to extract non-setback ordinance values from text.

    Parameters
    ----------
    **kwargs
        Keyword-value pairs to add to graph.

    Returns
    -------
    nx.DiGraph
        Graph instance that can be used to initialize an
        `elm.tree.DecisionTree`.
    """
    G = _setup_graph_no_nodes(**kwargs)

    G.add_node(
        "init",
        prompt=(
            "We will attempt to extract structured data for this "
            "ordinance. Let's think step by step. Does the following text "
            "explicitly limit the {restriction} allowed for {wes_type}? "
            "Do not infer based on other restrictions; if this particular "
            "restriction is not explicitly mentioned then say 'No'. Pay extra "
            "attention to clarifying text found in parentheses and footnotes. "
            "Begin your response with either 'Yes' or 'No' and explain "
            "your answer."
            '\n\n"""\n{text}\n"""'
        ),
    )
    G.add_edge("init", "final", condition=llm_response_starts_with_yes)

    G.add_node(
        "final",
        prompt=(
            "Now we are ready to extract structured data. Respond based "
            "on our entire conversation so far. Return your answer in JSON "
            "format (not markdown). Your JSON file must include exactly four "
            'keys. The keys are "value", "units", "section", "comment". The '
            'value of the "value" key should be a numerical value '
            "corresponding to the {restriction} allowed for {wes_type}, or "
            "`null` if the text does not mention such a restriction. Use our "
            'conversation to fill out this value. The value of the "units" '
            "key should be a string corresponding to the units for the "
            "{restriction} allowed for {wes_type} by the text below, or "
            "`null` if the text does not mention such a restriction. Make "
            'sure to include any "per XXX" clauses in the units. '
            "{SECTION_PROMPT} {COMMENT_PROMPT}"
        ),
    )
    return G



# -*- coding: utf-8 -*-
"""
ELM decision trees.
"""
import networkx as nx
import logging


logger = logging.getLogger(__name__)


class DecisionTree:
    """Class to traverse a directed graph of LLM prompts. Nodes are
    prompts and edges are transitions between prompts based on conditions
    being met in the LLM response."""

    def __init__(self, graph):
        """Class to traverse a directed graph of LLM prompts. Nodes are
        prompts and edges are transitions between prompts based on conditions
        being met in the LLM response.

        Examples
        --------
        Here's a simple example to setup a decision tree graph and run with the
        DecisionTree class:

        >>> import logging
        >>> import networkx as nx
        >>> from rex import init_logger
        >>> from elm.base import ApiBase
        >>> from elm.tree import DecisionTree
        >>>
        >>> init_logger('elm.tree')
        >>>
        >>> G = nx.DiGraph(text='hello', name='Grant',
                           api=ApiBase(model='gpt-35-turbo'))
        >>>
        >>> G.add_node('init', prompt='Say {text} to {name}')
        >>> G.add_edge('init', 'next', condition=lambda x: 'Grant' in x)
        >>> G.add_node('next', prompt='How are you?')
        >>>
        >>> tree = DecisionTree(G)
        >>> out = tree.run()
        >>>
        >>> print(tree.all_messages_txt)

        Parameters
        ----------
        graph : nx.DiGraph
            Directed acyclic graph where nodes are LLM prompts and edges are
            logical transitions based on the response. Must have high-level
            graph attribute "api" which is an ApiBase instance. Nodes should
            have attribute "prompt" which is a string that can have {format}
            named arguments that will be filled from the high-level graph
            attributes. Nodes can also have "callback" attributes that are
            callables that act on the LLM response in an arbitrary way. The
            function signature for a callback must be
            ``callback(llm_response, decision_tree, node_name)``.
            Edges can have attribute "condition" that is a callable to be
            executed on the LLM response text that determines the edge
            transition. An edge from a node without a condition acts as an
            "else" statement if no other edge conditions are satisfied. A
            single edge from node to node does not need a condition.
        """
        self._g = graph
        self._history = []
        assert isinstance(self.graph, nx.DiGraph)
        assert 'api' in self.graph.graph

    def __getitem__(self, key):
        """Retrieve a node by name (str) or edge by (node0, node1) tuple"""
        out = None
        if key in self.graph.nodes:
            out = self.graph.nodes[key]
        elif key in self.graph.edges:
            out = self.graph.edges[key]
        else:
            msg = (f'Could not find "{key}" in graph')
            logger.error(msg)
            raise KeyError(msg)
        return out

    @property
    def api(self):
        """Get the ApiBase object.

        Returns
        -------
        ApiBase
        """
        return self.graph.graph['api']

    @property
    def messages(self):
        """Get a list of the conversation messages with the LLM.

        Returns
        -------
        list
        """
        return self.api.messages

    @property
    def all_messages_txt(self):
        """Get a printout of the full conversation with the LLM

        Returns
        -------
        str
        """
        return self.api.all_messages_txt

    @property
    def history(self):
        """Get a record of the nodes traversed in the tree

        Returns
        -------
        list
        """
        return self._history

    @property
    def graph(self):
        """Get the networkx graph object

        Returns
        -------
        nx.DiGraph
        """
        return self._g

    def call_node(self, node_name):
        """Call the LLM with the prompt from the input node and search the
        successor edges for a valid transition condition

        Parameters
        ----------
        node_name : str
            Name of node being executed.

        Returns
        -------
        out : str
            Next node or LLM response if at a leaf node.
        """

        node = self[node_name]
        prompt = self._prepare_graph_call(node_name)
        out = self.api.chat(prompt)
        node['response'] = out

        if 'callback' in node:
            callback = node['callback']
            callback(out, self, node_name)

        return self._parse_graph_output(node_name, out)

    def _prepare_graph_call(self, node_name):
        """Prepare a graph call for given node."""
        prompt = self[node_name]['prompt']
        txt_fmt = {k: v for k, v in self.graph.graph.items() if k != 'api'}
        prompt = prompt.format(**txt_fmt)
        self._history.append(node_name)
        return prompt

    def _parse_graph_output(self, node0, out):
        """Parse graph output for given node and LLM call output. """
        successors = list(self.graph.successors(node0))
        edges = [self[(node0, node1)] for node1 in successors]
        conditions = [edge.get('condition', None) for edge in edges]

        if len(successors) == 0:
            logger.info(f'Reached leaf node "{node0}".')
            return out

        if len(successors) > 1 and all(c is None for c in conditions):
            msg = (f'At least one of the edges from "{node0}" should have '
                   f'a "condition": {edges}')
            logger.error(msg)
            raise AttributeError(msg)

        # prioritize callable conditions
        for i, condition in enumerate(conditions):
            if callable(condition) and condition(out):
                logger.info(f'Node transition: "{node0}" -> "{successors[i]}" '
                            '(satisfied by callable condition)')
                return successors[i]

        # None condition is basically "else" statement
        for i, condition in enumerate(conditions):
            if condition is None:
                logger.info(f'Node transition: "{node0}" -> "{successors[i]}" '
                            '(satisfied by None condition)')
                return successors[i]

        msg = (f'None of the edge conditions from "{node0}" '
               f'were satisfied: {edges}')
        logger.error(msg)
        raise AttributeError(msg)

    def run(self, node0='init'):
        """Traverse the decision tree starting at the input node.

        Parameters
        ----------
        node0 : str
            Name of starting node in the graph. This is typically called "init"

        Returns
        -------
        out : str
            Final response from LLM at the leaf node.
        """

        self._history = []

        while True:
            try:
                out = self.call_node(node0)
            except Exception as e:
                last_message = self.messages[-1]['content']
                msg = ('Ran into an exception when traversing tree. '
                       'Last message from LLM is printed below. '
                       'See debug logs for more detail. '
                       '\nLast message: \n'
                       f'"""\n{last_message}\n"""')
                logger.debug('Error traversing trees, heres the full '
                             'conversation printout:'
                             f'\n{self.all_messages_txt}')
                logger.error(msg)
                raise RuntimeError(msg) from e
            if out in self.graph:
                node0 = out
            else:
                break

        logger.info(f'Output: {out}')

        return out



# -*- coding: utf-8 -*-
"""ELM Web Scraping - DuckDuckGo search"""
import random
import asyncio
import logging

from duckduckgo_search import DDGS

from elm.web.search.base import (PlaywrightSearchEngineLinkSearch,
                                 SearchEngineLinkSearch)


logger = logging.getLogger(__name__)


_DDGS_SEMAPHORE = asyncio.Semaphore(1)


class PlaywrightDuckDuckGoLinkSearch(PlaywrightSearchEngineLinkSearch):
    """Search for top links on the main DuckDuckGo search engine"""

    MAX_RESULTS_CONSIDERED_PER_PAGE = 10
    """Number of results displayed per DuckDuckGo page"""

    _SE_NAME = "DuckDuckGo"
    _SE_URL = "https://duckduckgo.com/"
    _SE_SR_TAG = '[data-testid="result-extras-url-link"]'
    _SE_QUERY_URL = (
        "https://duckduckgo.com/?q={}&kl=us-en&kc=-1&kz=-1&kaf=1&ia=web"
    )

    async def _perform_homepage_search(self, page, search_query):
        """Fill in search bar with user query and hit enter"""
        await self._move_mouse(page)

        logger.trace("Clicking on search bar")

        search_bar = page.locator('#searchbox_input')
        await self._move_and_click(page, search_bar)
        await asyncio.sleep(random.uniform(0.5, 1.5))

        logger.trace("Typing in query: %r", search_query)
        await page.keyboard.type(search_query, delay=random.randint(80, 150))
        await asyncio.sleep(random.uniform(0.5, 1.5))

        logger.trace("Hitting enter for query: %r", search_query)
        await page.keyboard.press("Enter")


class APIDuckDuckGoSearch(SearchEngineLinkSearch):
    """Search the web for links using the DuckDuckGo API"""

    _SE_NAME = "DuckDuckGo API"

    def __init__(self, region="wt-wt", backend="auto", timeout=10,
                 verify=True, sleep_min_seconds=10, sleep_max_seconds=20):
        """

        Parameters
        ----------
        region : str, optional
            DDG search region param. By default, ``"wt-wt"``, which
            signifies no region.
        backend : {auto, html, lite}, optional
            Option for DDG search type.

                - auto: select randomly between HTML and Lite backends
                - html: collect data from https://html.duckduckgo.com
                - lite: collect data from https://lite.duckduckgo.com

            By default, ``"auto"``.
        timeout : int, optional
            Timeout for HTTP requests, in seconds. By default, ``10``.
        verify : bool, optional
            Apply SSL verification when making the request.
            By default, ``True``.
        sleep_min_seconds : int, optional
            Minimum number of seconds to sleep between queries. We
            recommend not setting this below ``5`` seconds to avoid
            rate limiting errors thrown by DuckDuckGo.
            By default, ``10``.
        sleep_max_seconds : int, optional
            Maximum number of seconds to sleep between queries.
            By default, ``20``.
        """
        self.region = region
        self.backend = backend
        self.timeout = timeout
        self.verify = verify
        self.sleep_min_seconds = sleep_min_seconds
        self.sleep_max_seconds = sleep_max_seconds

    async def _search(self, query, num_results=10):
        """Search web for links related to a query"""

        ddgs = DDGS(timeout=self.timeout, verify=self.verify)
        results = ddgs.text(query, region=self.region,
                            backend=self.backend,
                            max_results=num_results)

        return list(filter(None, (info.get('href', "").replace("+", "%20")
                                  for info in results)))

    async def _skip_exc_search(self, query, num_results=10):
        """Sleep between DDG searched to avoid rate limiting"""
        async with _DDGS_SEMAPHORE:
            try:
                out = await self._search(query, num_results=num_results)
            except Exception as e:
                logger.exception(e)
                out = []

            await self._sleep_after_query(query)

        return out

    async def _sleep_after_query(self, query):
        """Sleep for a random time after a query"""
        delay = random.uniform(self.sleep_min_seconds, self.sleep_max_seconds)
        logger.debug("DDG search sleeping for %.2f seconds after query: %s",
                     delay, query)
        await asyncio.sleep(delay)



# -*- coding: utf-8 -*-
"""
Energy Language Model
"""

import os
import logging
from functools import partial, partialmethod
from elm.base import ApiBase
from elm.chunk import Chunker
from elm.embed import ChunkAndEmbed
from elm.pdf import PDFtoTXT
from elm.summary import Summary
from elm.tree import DecisionTree
from elm.wizard import EnergyWizard
from elm.web.osti import OstiRecord, OstiList

__author__ = """Grant Buster"""
__email__ = "Grant.Buster@nrel.gov"

ELM_DIR = os.path.dirname(os.path.realpath(__file__))
TEST_DATA_DIR = os.path.join(os.path.dirname(ELM_DIR), 'tests', 'data')

logging.TRACE = 5
logging.addLevelName(logging.TRACE, 'TRACE')
logging.Logger.trace = partialmethod(logging.Logger.log, logging.TRACE)
logging.trace = partial(logging.log, logging.TRACE)
